"""
Command generate-workflow
"""
import re
import sys

import click
import yaml
from tcd_pipeline import pipeline as pipeline_tool


@click.command(help="Generate workflows from pipelines")
@click.option("--pipelines",
  default=".",
  type=click.Path(),
  show_default=True,
  help="Path to pipeline file or directory contains *-pipelines.yaml"
  )
@click.option("--git-ref-type",
  help="Git ref type",
  type=click.Choice(["branch", "tag"])
  )
@click.option("--git-ref-name",
  help="Branch name or Tag name"
  )
@click.option("--change",
  help="Changed file(s). Example Dockerfile  or [Dockerfile,README.md]",
  multiple=True,
  default=[list]
  )
@click.option("--output",
  help="Output workflows file. [Default: stdout]",
  type=click.Path()
  )
def generate_workflows(pipelines: str,
                      git_ref_type: str,
                      git_ref_name: str,
                      change: list[str],
                      output: str):
    """Generate workflows from pipelines"""
    if output == "" or output == "stdout":
        output = None

    fact: dict = yaml.safe_load(f"""
      refType: {git_ref_type}
      refName: {git_ref_name}
      changes: []
    """)

    # Add change to fact
    for file in change:
        if file.__contains__(","):
            for item_file in re.sub(r"[\[\]]", "", file).split(","):
                fact.get("changes").append(item_file)
        else:
            fact.get("changes").append(file)

    try:
        pipeline_tool.generate_workflows(
          pipeline_path=pipelines,
          event=fact,
          output=output
        )
    except Exception as err:
        print(err)
        sys.exit(1)



"""Evaluate list of changes matched with rule glob pattern

Examples
--------
glob_patterns : list[str]
    List of pattern to evaluate
facts : list[str]
    List of changes
"""
from fnmatch import fnmatch


def is_changes_match_glob(glob_patterns: list[str], facts: list[str]) -> bool:
    """Validate fact match at lease one glob pattern

    This function will be return True when atleast  pattern match with atleast fact

    Parameters
    ----------
    glob_patterns : list[str]
        List of glob pattern to be verified
    facts : list[str]
        Fact to be verified

    Returns
    -------
    bool
        True if atleast item matched
    """

    if not glob_patterns:
        return True

    for pattern in glob_patterns:
        for fact in facts:
            if fnmatch(fact, pattern):
                return True

    return False



"""Logging configuration

This module contains functions for setup logging with yaml file.
"""
import importlib.resources as pkg_resources
import logging
import logging.config
import os

import yaml
from tcd_pipeline import config as local_config


def config_logging(
  default_config: str = 'logging.yaml',
  default_level: int = logging.INFO,
  env_key: str = 'LOG_CFG') -> None:
    """Setup logging configuration from config file or dictionary file

    Ref https://gist.github.com/kingspp/9451566a5555fb022215ca2b7b802f19

    By default, this module will be loading resources/logging.yaml
    to config logging.

    This can be overwrite by environment variable LOG_CFG
    to config file (*.yaml)

    Parameters
    ----------
    default_config : str, optional
      Default internal yaml logging configuration. Default is logging.yaml
    default_level : int
      Default log level. Default value is "logging.INFO"
    env_key : str
      Environment variable name to logging configuration file (yaml)

    Returns
    -------
    None
    """

    # path = default_path
    # LOG_CFG is provided. Use user-defined configuration file
    path = os.getenv(env_key, None)
    if path:
        if os.path.exists(path):
            with open(path, 'rt', encoding="UTF-8") as config_file:
                try:
                    config = yaml.safe_load(config_file.read())
                    logging.config.dictConfig(config)
                except IOError as err:
                    print(err)
                    print(f"""Error in provided logging configuration [{path}].
                      Using default configs""")
                    logging.basicConfig(level=default_level)
        else:
            logging.basicConfig(level=default_level)
            print(f"""Failed to load configuration file {path}.
              Using default configs""")

    # LOG_CFG is not provided. Use default in package tcd_config.logging.yaml.
    else:
        try:
            pkg_resources.open_text(local_config, default_config)
            config = yaml.safe_load(pkg_resources.open_text(
              local_config, default_config))
            logging.config.dictConfig(config)
        except IOError as err:
            print(err)
            print(f"""Error in default configuration [{default_config}].
              Using default configs""")
            logging.basicConfig(level=default_level)

    return None



"""Evaluate push event with push rules

Examples
--------
    event : dict
        refType: branch
        refName: release/dev
        changes:
          - Dockerfilex
          - src/main/java/Main.java

    rules : list[dict]
        when:
          push:
            - type: branch # [branch, tag]
              matcher: exact # [ regex, glob, semver, exact] https://semver.org/#is-v123-a-semantic-version
              value: master  #
            - type: branch
              matcher: glob
              value: release/*
              changes: # When one of this file list changed.
                - "Dockerfile"
                - "src/main/java/**.java"
            - type: tag
              matcher: semver
            - type: tag
              matcher: regex
              value: release/.+
"""
import re
from fnmatch import fnmatch

import tcd_pipeline.rules.changes as changes_rules
from jsonpath_ng.ext import parse


def is_git_ref_match_glob(rules: list[dict], fact: dict) -> bool:
    """Evaluate fact matched with glob

    Parameters
    ----------
    rule : list[dict]
        List of rule.

        Example: list of rules
        [
          {
            'type': 'branch',
            'matcher': 'glob',
            'value': 'release/*',
            'changes': [
              'Dockerfile',
              'src/main/java/**.java'
            ]
          }
        ]
    fact : dict
        Git push event data

        Example:
        {
          'refType', 'branch',
          'refName', 'release/dev',
          'changes', [
            'Dockerfile',
            'src/main/java/Main.java'
          ]

        }
    """

    actual: str = fact.get("refName")
    fact_ref_type: str = fact.get("refType")
    jsonpath_expr = parse(f"$..push[?((@.type=='{fact_ref_type}')&(@.matcher=='glob'))]")
    for rule in jsonpath_expr.find(rules):
        if fnmatch(actual, rule.value.get("value")):
            return changes_rules.is_changes_match_glob(
              glob_patterns=rule.value.get("changes"),
              facts=fact.get("changes")
            )

    return False


def is_git_ref_match_regex(rules: list[dict], fact: dict) -> bool:
    """Evaluate fact matched with regex

    Parameters
    ----------
    rule : list[dict]
        List of rule.

        Example: list of rules
        [
          {
            'type': 'branch',
            'matcher': 'regex',
            'value': 'release/.+',
            'changes': [
              'Dockerfile',
              'src/main/java/**.java'
            ]
          }
        ]
    fact : dict
        Git push event data

        Example:
        {
          'refType', 'branch',
          'refName', 'release/dev',
          'changes', [
            'Dockerfile',
            'src/main/java/Main.java'
          ]
        }
    """

    actual: str = fact.get("refName")
    fact_ref_type: str = fact.get("refType")
    jsonpath_expr = parse(f"$..push[?((@.type=='{fact_ref_type}')&(@.matcher=='regex'))]")
    for rule in jsonpath_expr.find(rules):
        if bool(re.match(pattern=rule.value.get("value"), string=actual)):
            return changes_rules.is_changes_match_glob(
              glob_patterns=rule.value.get("changes"),
              facts=fact.get("changes")
            )

    return False


def is_git_ref_match_semver(rules: list[dict], fact: dict) -> bool:
    """Evaluate fact matched with semver

    Parameters
    ----------
    rule : list[dict]
        List of rule.

        Example: list of rules
        [
          {
            'type': 'branch',
            'matcher': 'semver'
            'value': '(?i)((v|version|release)[\.\-\/]?)?', # prefix
            'changes': [
              'Dockerfile',
              'src/main/java/**.java'
            ]
          }
        ]
    fact : dict
        Git push event data

        Example:
        {
          'refType', 'branch',
          'refName', 'release/dev',
          'changes', [
            'Dockerfile',
            'src/main/java/Main.java'
          ]

        }
    """

    actual: str = fact.get("refName")
    fact_ref_type: str = fact.get("refType")
    jsonpath_expr = parse(f"$..push[?((@.type=='{fact_ref_type}')&(@.matcher=='semver'))]")
    for rule in jsonpath_expr.find(rules):
        # default semver prefix
        prefix_regex_pattern = '(?i)((v|version|release)[\.\-\/]?)?'
        semver_regex_pattern = '\d+(\.\d+){0,2}(\-.+)?'

        # Replace default prefix from rule value
        if rule.value.get("value"):
            prefix_regex_pattern = rule.value.get("value")
        regex_pattern = f"{prefix_regex_pattern}{semver_regex_pattern}"

        if bool(re.match(pattern=regex_pattern, string=actual)):
            return changes_rules.is_changes_match_glob(
              glob_patterns=rule.value.get("changes"),
              facts=fact.get("changes")
            )

    return False


def is_git_ref_match_exact(rules: list[dict], fact: dict) -> bool:
    """Evaluate fact matched with semver

    Parameters
    ----------
    rule : list[dict]
        List of rule.

        Example: list of rules
        [
          {
            'type': 'branch',
            'matcher': 'exact',
            'value': 'develop',
            'changes': [
              'Dockerfile',
              'src/main/java/**.java'
            ]
          }
        ]
    fact : dict
        Git push event data

        Example:
        {
          'refType', 'branch',
          'refName', 'develop',
          'changes', [
            'Dockerfile',
            'src/main/java/Main.java'
          ]

        }
    """

    actual: str = fact.get("refName")
    fact_ref_type: str = fact.get("refType")
    jsonpath_expr = parse(f"$..push[?((@.type=='{fact_ref_type}')&(@.matcher=='exact'))]")
    for rule in jsonpath_expr.find(rules):
        if rule.value.get("value") == actual:
            return changes_rules.is_changes_match_glob(
              glob_patterns=rule.value.get("changes"),
              facts=fact.get("changes")
            )

    return False



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tcd_pipeline',
 'tcd_pipeline.commands',
 'tcd_pipeline.config',
 'tcd_pipeline.rules',
 'tcd_pipeline.templates']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'click>=8.1.3,<9.0.0', 'jsonpath-ng>=1.5.3,<2.0.0']

entry_points = \
{'console_scripts': ['tcd-pipeline = tcd_pipeline.__main__:cli']}

setup_kwargs = {
    'name': 'tcd-pipeline',
    'version': '0.2.4',
    'description': '',
    'long_description': '# TCD pipeline tool\n\nCLI tool for convert TCD pipelines to ArgoWrokflows manifests.\n\n---\n\n## Build\n\nBuild instruction can be found at [BUILD.md](BUILD.md)\n\n---\n\n## Release\n\nRelease instruction and history can be found at [RELEASE.md](RELEASE.md)\n\n---\n\n## Install\n\n```sh\npip install tcd-pipeline\n```\n\n---\n\n## Usage\n\nRunning from installed\n\n```sh\ntcd-pipeline --help\n```\n\nRunning with python module\n\n```sh\npython -m tcd_pipeline\n```\n',
    'author': 'chitchai',
    'author_email': 'chitchaic@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



""""Include local template

"""


def get_templates(alias: str, templates: list[str]):
    """Collect reference templates from files

    Parameters
    ----------
    alias : str
        Alias name will be patched as prefix of include template

    """
    pass



"""Main CLI application"""

import click

from tcd_pipeline.commands import generate_workflows


@click.group(help="TCD Pipeline tools")
@click.version_option(package_name="tcd_pipeline", message="version %(version)s")
def cli():
    """Main cli function"""


cli.add_command(generate_workflows.generate_workflows)

if __name__ == '__main__':
    cli()
